\name{MCPDT}

\alias{MCPDT}
\alias{MCPPAT}
\alias{MCPPATu}

\title{
  Monte Carlo Pedigree Disequilibrium Test (MCPDT) and
  Monte Carlo Pedigree Parental-Asymmetry Test (MCPPAT)
  Monte Carlo Pedigree Parental-Asymmetry Test utilizing both affected and unaffected offspring(MCPPATu)}

\description{
  Pedigree disequilibrium test (PDT), pedigree parental-asymmetry
  test (PPAT), pedigree parental-asymmetry test utilizing both affected and unaffected offspring(PPATu) with missing genotypes simulated
  conditioned on observed genotypes.}

\usage{
  MCPDT(ped,mcsize,loci=NULL,af=NULL,sexlinked=FALSE)

  MCPPAT(ped,mcsize,loci=NULL,af=NULL,sexlinked=FALSE)
  
  MCPPATu(ped,mcsize,loci=NULL,af=NULL,sexlinked=FALSE)}

\arguments{
  \item{ped}{name of a standard LINKAGE pedigree file (pre-makeped
    format) or a matrix/dataframe containing the pedigree data}
  \item{mcsize}{Monte Carlo sample size}
  \item{loci}{name of a standard LINKAGE loci file}
  \item{af}{matrix of marker allele frequencies, one column for each marker}
  \item{sexlinked}{whether the markers are on X chromosome}
}

\details{
  \sQuote{MCPDT} calculates PDT and MCPDT statistics and p-values.
  
  \sQuote{MCPPAT} calculates PDT, MCPDT, PPAT and MCPPAT statistics and
  p-values.

\sQuote{MCPPATu} calculates PDT, MCPDT, PPAT, MCPPAT,PPATu and MCPPATu statistics and
  p-values.  

  \var{ped} is required. Only SNP markers are supported.
  If loci file is provided, allele frequencies are read
  from the file specified. Otherwise, they can be provided directly in the
  argument af. If allele frequencies are not provided in either way, they
  are estimated from all the typed founders' genotypes.

  If there are multiple markers, each marker is analyzed separately.
  
  \var{sexlinked} only applies to MCPDT since MCPPAT only deals with
  autosomal markers. If \var{sexlinked}=TRUE in \sQuote{MCPPAT},
  only PDT and MCPDT will be calculated.
}
\value{
  \item{Tstat}{PDT T statistics for the markers}
  \item{pvalue}{P values assuming asymptotic normal distribution}
  \item{mcsize}{Monte Carlo sample size}
  \item{af}{allele frequencies used in the simulation}
}
\references{
  Martin, E. R., Monks, S. A., Warren, L. L. and Kaplan, N. L. (2000) A
  test for linkage and association in general pedigrees: the pedigree
  disequilibrium test. Am J Hum Genet 67:146-154

  Ding, J., Lin, S. and Liu, Y. (2006) Monte Carlo pedigree
  disequilibrium test for markers on the X chromosome. Am J Hum Genet
  79:567-573  

  Zhou, J.Y., Ding, J., Fung, W.K. and Lin, S. (2009) Detection of
  parent-of-origin effects using general pedigree data. Submitted
  
  Ott, J. (1989) Computer-simulation methods in human linkage analysis.
  Proc Natl Acad Sci USA 86:4175-4178

  Weeks, D. E., Ott, J., and Lathrop, G. M. (1990) SLINK: A general
  simulation program for linkage analysis. Am J Hum Genet 47:A204 
}
\examples{
}
\keyword{htest}

