/*This file is part of a C version of SLINK a simulation program based
on the LINKAGE package. SLINK was originally described in: D. E.
Weeks, J. Ott, G. M. Lathrop (1990), SLINK: a general simulation
program for linkage analysis, American Journal of Human Genetics
47(1990), A204 (abstract). This version of SLINK is adapted to use the
faster pedigree traversal algorithms described in: R. W. Cottingham
Jr., R. M. Idury, A. A. Schaffer (1993), Faster sequential genetic
linkage computations, American Journal of Human Genetics 53(1993), pp.
252-263.*/
/*This file contains definitions and declaration, which differ from those
used in the LINKAGE programs. In some cases the declarations are totally
new and in other cases they are changed*/
/* 8 December 1993, A. A. Schaffer pulled these out of slink.c and adapted
for faster algorithms */

#define print           false
    /*FALSE turns off some output for quicker running*/
#define feedback        false   /*DESCRIPTION OF PATH THROUGH FAMILIES*/

#define bufsize         4096
    /*Buffer size in bytes for PEDFILE output file*/

#define unafqu          1
/*CODE FOR UNAFFECTED MALES AT A SEX-LINKED
  QUANTITATIVE TRAIT; This must not equal affall*/
/* AFFECTION STATUS */
#define unaff           1   /*CODE FOR UNAFFECTED INDIVIDUAL*/
/*Note: unaff must not equal affval */
    /*MAXIMUM NUMBER OF BINARY CODES AT A SINGLE LOCUS*/
/* OTHERS */

#define scale           1.0   /*SCALE FACTOR*/
#define zerolike        (-1.0e20)

typedef char hapvector[maxlocus];

typedef struct locusvalues {
  long nallele, format;
  double freq[maxall];
  struct locusvalues *privlocus;
  locustype which;
  union {
    struct {
      double pen[maxall + 1][maxall][3][maxliab];
      long nclass;
    } U0;
    struct {
      long ntrait;
      means pm;
      covmatrix vmat;
      double det, contrait, conmat;
    } U1;
    struct {
      phenarray allele;
      long numfact;
    } U2;
  } UU;
} locusvalues;


typedef struct thisperson {
  long origped;   /* L. Sandkuijl's suggestion */
  long id, ped, inloop, geneloc, prbnum, availnum, origid;
  /*geneloc = location of the simulated genotype in the genarray;
  prbnum = proband number; availnum = availability code;
  origid = original id found in pedigree file*/
  struct thisperson *pa, *ma, *foff, *nextpa, *nextma;
  thisarray *gen;
  indphen phen, phen1, phen2;
  phenotype *privphen;
  boolean unknown, multi, done, up, male, firstpass;
  information *store;
} thisperson;


double wsec, wfsec, isec, fsec, hetvalue, finish, inc;
/* Used for timing replicates */
/*SUN*/
long code[4];   /* Counter of availnum codes */
long men, women;   /* Counters of men and women in simped.dat */
boolean noloop;
genotype geneprob;
    /*Array of genotype probabilities taken from genarray*/
long totall, trait, replicates;
/*trait = number of the trait locus;
replicates = desired number of replications */
long numgen, seed1, seed2, seed3;
/*numgen = number of genotypes; random number seeds*/
char hapstore[maxlocus][maxhap];
/*Array of alleles corresponding to the position geneloc in genarray*/
boolean zeromale[maxlocus], zerofemale[maxlocus];
locusvalues *thislocus[maxlocus], *thislocus1[maxlocus], *thislocus2[maxlocus];
long order1[maxlocus], order2[maxlocus];
/*PEOPLE*/
thisperson *person[maxind + 1];
thisperson *proband[maxped];
thisperson *looppers[maxped][maxloop][2];
/*OTHERS*/
long j, nlocus, which, thissystem;
boolean  mapping, unlinked, linked;
FILE *simout, *simped, *simdata, *pedfile;
/*pedfile = where the simulated pedigrees are written to*/
long k;


/* Local variables for seg: */
struct LOC_seg {
  struct LOC_likelihood *LINK;
  thisperson **p, **q, **r, *child, *father, *mother;
  long fseg, sseg, sstart, send, fstart, fend, nfirst, nsecond, firstseg,
       secondseg;
  double pf, ps;
  thetavalues *firstsex, *secondsex;
} ;

/*Local variables for likelihood*/
struct LOC_likelihood {
  long thisped;
  thisperson *proband;
  long loopgen[maxloop];
  double homo, hetero;
  long nuscale;
  thisarray *holdpoint[maxloop];
} ;

//for interface between R and C
struct Rinput {
  int nind, nlocus;
  int **ped;
  int autosomal;
  double *af; //allele frequency
  int seed; //random seed
  int replicates; //number of replicate
  int trait; //position of trait locus
  int hetvalue; //heterogeneity percentage
};

struct Rinput r1; 

struct MCoutput {
  int mcsize,nind,nped,nlocus;
  int **ped;
  double **Dstat;
};

struct MCoutput mc1; 

