/* This file is part of a C version of SLINK a simulation program based
on the LINKAGE package. SLINK was originally described in: D. E.
Weeks, J. Ott, G. M. Lathrop (1990), SLINK: a general simulation
program for linkage analysis, American Journal of Human Genetics
47(1990), A204 (abstract). This version of SLINK is adapted to use the
faster pedigree traversal algorithms described in: R. W. Cottingham
Jr., R. M. Idury, A. A. Schaffer (1993), Faster sequential genetic
linkage computations, American Journal of Human Genetics 53(1993), pp.
252-263.*/
/*This file contains many of the I/O routines*/
/* December 1993 Sliced out of slink.c by A. A. Schaffer*/

#include <stdio.h>
#include "commondefs.h"
#include "sldefs.h"


/* Local variables for recombination: */
struct LOC_recombination {
  long here;
  double p1, p2, p3, p4;
} ;

/* Local variables for recombine: */
struct LOC_recombine {
  struct LOC_recombination *LINK;
  double *theta;
  double *segprob;
  long there;
  long nhap;
  boolean thishet[maxlocus];
  hapvector hap1, hap2;
  hapvector thishap1[maxseg], thishap2[maxseg];
} ;

static void scramble(LINK)
struct LOC_recombine *LINK;
{
  long whichhap, start, length, i, j, k, FORLIM2;
  double recval, val;

  start = 0;
  do {
    start++;
  } while (!LINK->thishet[start - 1]);
  length = LINK->there - LINK->LINK->here;
  for (i = 1; i < length; i++) {
    memcpy(LINK->hap1, LINK->thishap1[i], sizeof(hapvector));
    for (j = 1; j <= length; j++) {
      val = 0.5;
      whichhap = 1;
      recval = LINK->theta[start - 1];
      FORLIM2 = nlocus;
      for (k = start; k < FORLIM2; k++) {
	if (!LINK->thishet[k])
	  recval = recval * (1.0 - LINK->theta[k]) +
		   LINK->theta[k] * (1.0 - recval);
	else {
	  if (whichhap == 1) {
	    if (LINK->thishap1[j - 1][k] == LINK->hap1[k])
	      val *= 1 - recval;
	    else {
	      val *= recval;
	      whichhap = 2;
	    }
	  } else {
	    if (LINK->thishap2[j - 1][k] == LINK->hap1[k])
	      val *= 1 - recval;
	    else {
	      val *= recval;
	      whichhap = 1;
	    }
	  }
	  recval = LINK->theta[k];
	}
      }
      LINK->there++;
      LINK->segprob[LINK->there - 1] = val;
    }
  }
}  /* scramble */


static void setrec(val, LINK)
double val;
struct LOC_recombine *LINK;
{
  LINK->nhap++;
  memcpy(LINK->thishap1[LINK->nhap - 1], LINK->hap1, sizeof(hapvector));
  memcpy(LINK->thishap2[LINK->nhap - 1], LINK->hap2, sizeof(hapvector));
  LINK->there++;
  LINK->segprob[LINK->there - 1] = val;
}  /* setrec */


static void dointer(LINK)
struct LOC_recombine *LINK;
{
  long i;
  boolean temphet[3];
  long FORLIM;

  FORLIM = nlocus;
  for (i = 0; i < FORLIM; i++)
    temphet[i] = LINK->thishet[i];

  if (temphet[0] && temphet[1] && !temphet[2]) {
    setrec(0.5 - 0.5 * LINK->theta[0], LINK);
    setrec(0.5 * LINK->theta[0], LINK);
    setrec(0.5 * LINK->theta[0], LINK);
    setrec(0.5 - 0.5 * LINK->theta[0], LINK);
    return;
  }
  if (temphet[2] && temphet[1] && !temphet[0]) {
    setrec(0.5 - 0.5 * LINK->theta[nlocus - 2], LINK);
    setrec(0.5 * LINK->theta[nlocus - 2], LINK);
    setrec(0.5 * LINK->theta[nlocus - 2], LINK);
    setrec(0.5 - 0.5 * LINK->theta[nlocus - 2], LINK);
    return;
  }
  if (temphet[2] && temphet[0] && !temphet[1]) {
    setrec(0.5 - 0.5 * LINK->theta[nlocus - 1], LINK);
    setrec(0.5 * LINK->theta[nlocus - 1], LINK);
    setrec(0.5 * LINK->theta[nlocus - 1], LINK);
    setrec(0.5 - 0.5 * LINK->theta[nlocus - 1], LINK);
    return;
  }
  if (!(temphet[0] && temphet[1] && temphet[2])) {
    setrec(0.5, LINK);
    return;
  }
  setrec(0.5 * LINK->LINK->p4, LINK);
  setrec(0.5 * LINK->LINK->p2, LINK);
  setrec(0.5 * LINK->LINK->p3, LINK);
  setrec(0.5 * LINK->LINK->p1, LINK);
  setrec(0.5 * LINK->LINK->p2, LINK);
  setrec(0.5 * LINK->LINK->p4, LINK);
  setrec(0.5 * LINK->LINK->p1, LINK);
  setrec(0.5 * LINK->LINK->p3, LINK);
  setrec(0.5 * LINK->LINK->p3, LINK);
  setrec(0.5 * LINK->LINK->p1, LINK);
  setrec(0.5 * LINK->LINK->p4, LINK);
  setrec(0.5 * LINK->LINK->p2, LINK);
  setrec(0.5 * LINK->LINK->p1, LINK);
  setrec(0.5 * LINK->LINK->p3, LINK);
  setrec(0.5 * LINK->LINK->p2, LINK);
  setrec(0.5 * LINK->LINK->p4, LINK);

  /*not informative*/
}  /* dointer */


static void nexthet(i, val, inphase, LINK)
long i;
double val;
boolean inphase;
struct LOC_recombine *LINK;
{
  double newval, recval;

  recval = LINK->theta[i - 1];
  do {
    i++;
    LINK->hap1[i - 1] = 0;
    LINK->hap2[i - 1] = 0;
    if (!LINK->thishet[i - 1] && i != nlocus)
      recval = recval * (1 - LINK->theta[i - 1]) +
	       (1 - recval) * LINK->theta[i - 1];
  } while (!(i == nlocus || LINK->thishet[i - 1]));
  if (i != nlocus) {
    if (inphase)
      newval = val * (1 - recval);
    else
      newval = val * recval;
    LINK->hap1[i - 1] = 1;
    LINK->hap2[i - 1] = 2;
    nexthet(i, newval, true, LINK);
    LINK->hap2[i - 1] = 1;
    LINK->hap1[i - 1] = 2;
    if (!inphase)
      newval = val * (1 - recval);
    else
      newval = val * recval;
    nexthet(i, newval, false, LINK);
    return;
  }
  if (!LINK->thishet[i - 1]) {
    setrec(val, LINK);
    return;
  }
  if (inphase)
    newval = val * (1 - recval);
  else
    newval = val * recval;
  LINK->hap1[i - 1] = 1;
  LINK->hap2[i - 1] = 2;
  setrec(newval, LINK);
  if (!inphase)
    newval = val * (1 - recval);
  else
    newval = val * recval;
  LINK->hap2[i - 1] = 1;
  LINK->hap1[i - 1] = 2;
  setrec(newval, LINK);
}  /* nexthet */


static void getrecprob(LINK)
struct LOC_recombine *LINK;
{
  long i;

  LINK->nhap = 0;
  LINK->there = LINK->LINK->here;
  i = 0;
  do {
    i++;
    if (LINK->thishet[i - 1]) {
      LINK->hap1[i - 1] = 1;
      LINK->hap2[i - 1] = 2;
    } else {
      LINK->hap1[i - 1] = 0;
      LINK->hap2[i - 1] = 0;
    }
  } while (!(LINK->thishet[i - 1] || i == nlocus));
  if (i == nlocus)
    setrec(0.5, LINK);
  else if (interfer)
    dointer(LINK);
  else
    nexthet(i, 0.5, true, LINK);
  if (LINK->nhap > 1 && !interfer)
    scramble(LINK);
  LINK->LINK->here = LINK->there;
}  /* getrecprob */


static void gethet(system, LINK)
long *system;
struct LOC_recombine *LINK;
{
  long newsystem;

  newsystem = *system + 1;
  LINK->thishet[*system - 1] = false;
  if (*system != nlocus)
    gethet(&newsystem, LINK);
  else
    getrecprob(LINK);
  LINK->thishet[*system - 1] = true;
  if (*system != nlocus)
    gethet(&newsystem, LINK);
  else
    getrecprob(LINK);
}  /* gethet */

static void recombine(theta_, segprob_, LINK)
double *theta_;
double *segprob_;
struct LOC_recombination *LINK;
{
  struct LOC_recombine V;
  long system;


  V.LINK = LINK;
  V.theta = theta_;
  V.segprob = segprob_;
  LINK->here = 0;
  system = 1;
  gethet(&system, &V);
}  /* recombine */


static void getfemaletheta(LINK)
struct LOC_recombination *LINK;
{
  double dist;
  long ntheta, i;

  if (interfer)
    ntheta = nlocus;
  else
    ntheta = nlocus - 1;
  for (i = 0; i < ntheta; i++) {
    dist = getdist(&maletheta->theta[i]) * distratio;
    femaletheta->theta[i] = invdist(&dist);
  }
}  /* getfemaletheta */


void recombination()
{
  struct LOC_recombination V;
  long i;
  thetarray oldtheta;
  thetavalues *WITH;
  long FORLIM;


  if (interfer) {
    WITH = maletheta;
    if (mapping)
      WITH->theta[nlocus - 1] = mapfunction(WITH->theta[nlocus - 2],
					    WITH->theta[0]);
    memcpy(oldtheta, WITH->theta, sizeof(thetarray));
    if (!mapping && !dolod) {
      FORLIM = nlocus;
      for (i = 0; i < FORLIM; i++)
	oldtheta[i] = 1 / (1 + exp(oldtheta[i]));
      WITH->theta[0] = oldtheta[0] + oldtheta[nlocus - 1];
      WITH->theta[nlocus - 2] = oldtheta[nlocus - 2] + oldtheta[nlocus - 1];
      WITH->theta[nlocus - 1] = oldtheta[0] + oldtheta[nlocus - 2];
      V.p1 = oldtheta[0];
      V.p2 = oldtheta[nlocus - 2];
      V.p3 = oldtheta[nlocus - 1];
      V.p4 = 1.0 - V.p1 - V.p2 - V.p3;
    } else {
      V.p1 = (WITH->theta[0] +
	      WITH->theta[nlocus - 1] - WITH->theta[nlocus - 2]) / 2.0;
      V.p2 = (WITH->theta[nlocus - 2] +
	      WITH->theta[nlocus - 1] - WITH->theta[0]) / 2.0;
      V.p3 = (WITH->theta[nlocus - 2] +
	      WITH->theta[0] - WITH->theta[nlocus - 1]) / 2.0;
      V.p4 = 1.0 - V.p1 - V.p2 - V.p3;
    }
    recombine(WITH->theta, WITH->segprob, &V);
  } else
    recombine(maletheta->theta, maletheta->segprob, &V);
  if (sexdif) {
    if (!readfemale) {
      if (interfer && !dolod) {
	WITH = maletheta;
	WITH->theta[0] = oldtheta[0] + oldtheta[nlocus - 1];
	WITH->theta[nlocus - 2] = oldtheta[nlocus - 2] + oldtheta[nlocus - 1];
	WITH->theta[nlocus - 1] = oldtheta[0] + oldtheta[nlocus - 2];
      }
      getfemaletheta(&V);
    }
    if (interfer) {
      WITH = femaletheta;
      if (mapping)
	WITH->theta[nlocus - 1] = mapfunction(WITH->theta[nlocus - 2],
					      WITH->theta[0]);
      if (readfemale && !mapping && !dolod) {
	memcpy(oldtheta, WITH->theta, sizeof(thetarray));
	FORLIM = nlocus;
	for (i = 0; i < FORLIM; i++)
	  oldtheta[i] = 1 / (1 + exp(oldtheta[i]));
	WITH->theta[0] = oldtheta[0] + oldtheta[nlocus - 1];
	WITH->theta[nlocus - 2] = oldtheta[nlocus - 2] + oldtheta[nlocus - 1];
	WITH->theta[nlocus - 1] = oldtheta[0] + oldtheta[nlocus - 2];
	V.p1 = oldtheta[0];
	V.p2 = oldtheta[nlocus - 2];
	V.p3 = oldtheta[nlocus - 1];
	V.p4 = 1.0 - V.p1 - V.p2 - V.p3;
      } else {
	V.p1 = (WITH->theta[0] +
		WITH->theta[nlocus - 1] - WITH->theta[nlocus - 2]) / 2.0;
	V.p2 = (WITH->theta[nlocus - 2] +
		WITH->theta[nlocus - 1] - WITH->theta[0]) / 2.0;
	V.p3 = (WITH->theta[nlocus - 2] +
		WITH->theta[0] - WITH->theta[nlocus - 1]) / 2.0;
	V.p4 = 1.0 - V.p1 - V.p2 - V.p3;
      }
      recombine(WITH->theta, WITH->segprob, &V);
    } else
      recombine(femaletheta->theta, femaletheta->segprob, &V);
  }
  if (firsttime) {
    /*Introduced by Ramana*/
    nuneed=V.here;
    //    if (V.here < maxneed)
    //      printf(" Maxneed may be reduced to %12ld\n", V.here);
  }
}  /* recombination */


/* Local variables for getlocations: */
struct LOC_getlocations {
  long ngene, nseg, here, there, start, nhet, thisseg;
  boolean rarepresent, riskhom, riskhet;
  hapvector hap1, hap2;
  boolean thishet[maxlocus];
} ;

static boolean checkrare(LINK)
struct LOC_getlocations *LINK;
{
  long i;
  boolean check;
  long FORLIM;
  locusvalues *WITH;

  check = false;
  FORLIM = nlocus;
  for (i = 0; i < FORLIM; i++) {
    if (nohom[i]) {
      WITH = thislocus[i];
      if (WITH->freq[LINK->hap1[i] - 1] < minfreq ||
	  WITH->freq[LINK->hap2[i] - 1] < minfreq)
	check = true;
    }
  }
  return check;
}  /* checkrare */


static void checkrisk(riskhet, riskhom, LINK)
boolean *riskhet, *riskhom;
struct LOC_getlocations *LINK;
{
  *riskhet = false;
  *riskhom = false;
  if (LINK->hap1[risksys - 1] == riskall && LINK->hap2[risksys - 1] == riskall)
    *riskhom = true;
  else if ((LINK->hap1[risksys - 1] != riskall &&
	    LINK->hap2[risksys - 1] == riskall) ||
	   (LINK->hap2[risksys - 1] != riskall &&
	    LINK->hap1[risksys - 1] == riskall))
    *riskhet = true;
}  /* checkrisk */


static long gethapn(hap, LINK)
char *hap;
struct LOC_getlocations *LINK;
{
  long i, n, FORLIM;

  n = 1;
  FORLIM = nlocus;
  for (i = 0; i < FORLIM; i++)
    n += increment[i] * (hap[i] - 1);
  return n;
}  /* gethapn */

/* Local variables for domalerisk: */
struct LOC_domalerisk {
  struct LOC_getlocations *LINK;
} ;

static void setrisk(LINK)
struct LOC_domalerisk *LINK;
{
  long n;

  n = gethapn(LINK->LINK->hap1, LINK->LINK);
  if (LINK->LINK->hap1[risksys - 1] == riskall)
    riskmale[n - 1] = true;
  else
    riskmale[n - 1] = false;
}  /* setrisk */


static void getriskhap(system, LINK)
long system;
struct LOC_domalerisk *LINK;
{
  long i;
  locusvalues *WITH;
  long FORLIM;

  WITH = thislocus[system - 1];
  FORLIM = WITH->nallele;
  for (i = 1; i <= FORLIM; i++) {
    LINK->LINK->hap1[system - 1] = i;
    if (system != nlocus)
      getriskhap(system + 1, LINK);
    else
      setrisk(LINK);
  }
}  /* getriskhap */


static void domalerisk(LINK)
struct LOC_getlocations *LINK;
{
  struct LOC_domalerisk V;


  V.LINK = LINK;
  getriskhap(1L, &V);
}  /* domalerisk */

/* Local variables for domutation: */
struct LOC_domutation {
  struct LOC_getlocations *LINK;
} ;

static void setmutation(LINK)
struct LOC_domutation *LINK;
{
  long i, n;

  n = gethapn(LINK->LINK->hap1, LINK->LINK);
  if (LINK->LINK->hap1[mutsys - 1] == thislocus[mutsys - 1]->nallele)
    muthap[n - 1] = n;
  else {
    i = LINK->LINK->hap1[mutsys - 1];
    LINK->LINK->hap1[mutsys - 1] = thislocus[mutsys - 1]->nallele;
    muthap[n - 1] = gethapn(LINK->LINK->hap1, LINK->LINK);
    LINK->LINK->hap1[mutsys - 1] = i;
  }
  hapstore[mutsys - 1][muthap[n - 1] - 1] = LINK->LINK->hap1[mutsys - 1];
}  /* setmutation */


static void getmuthap(system, LINK)
long system;
struct LOC_domutation *LINK;
{
  long i;
  locusvalues *WITH;
  long FORLIM;

  WITH = thislocus[system - 1];
  FORLIM = WITH->nallele;
  for (i = 1; i <= FORLIM; i++) {
    LINK->LINK->hap1[system - 1] = i;
    if (system != nlocus)
      getmuthap(system + 1, LINK);
    else
      setmutation(LINK);
  }
}  /* getmuthap */


static void domutation(LINK)
struct LOC_getlocations *LINK;
{
  struct LOC_domutation V;


  V.LINK = LINK;
  getmuthap(1L, &V);
}  /* domutation */


static void setnumbers(LINK)
struct LOC_getlocations *LINK;
{
  long nhap1, nhap2, isys, FORLIM;

  LINK->ngene++;

  segstart[LINK->ngene - 1] = LINK->here + 1;
  probstart[LINK->ngene - 1] = LINK->there + 1;
  probend[LINK->ngene - 1] = LINK->there + LINK->nseg;

  LINK->there += LINK->nseg;

  nhap1 = gethapn(LINK->hap1, LINK);
  nhap2 = gethapn(LINK->hap2, LINK);
  FORLIM = nlocus;
  for (isys = 0; isys < FORLIM; isys++) {
    hapstore[isys][nhap1 - 1] = LINK->hap1[isys];
    hapstore[isys][nhap2 - 1] = LINK->hap2[isys];
  }
  gennustruct->genenumber[nhap1 - 1][nhap2 - 1] = LINK->ngene;
  gennustruct->genenumber[nhap2 - 1][nhap1 - 1] = LINK->ngene;

  if (minfreq != 0.0) {
    if (LINK->rarepresent)
      rare[LINK->ngene - 1] = true;
    else
      rare[LINK->ngene - 1] = false;
  } else
    rare[LINK->ngene - 1] = false;
  if (risk) {
    risk1[LINK->ngene - 1] = LINK->riskhet;
    risk2[LINK->ngene - 1] = LINK->riskhom;
  }

  LINK->thisseg++;
  seghap1[LINK->thisseg - 1] = nhap1;
  seghap2[LINK->thisseg - 1] = nhap2;
}  /* setnumbers */


static void hapscr(system, nscramble, LINK)
long system, nscramble;
struct LOC_getlocations *LINK;
{
  long i, j;

  if (LINK->thishet[system - 1])
    nscramble++;
  if (system != nlocus)
    hapscr(system + 1, nscramble, LINK);
  else
    setnumbers(LINK);
  if (nscramble <= 1)
    return;
  if (LINK->hap1[system - 1] == LINK->hap2[system - 1])
    return;
  i = LINK->hap1[system - 1];
  j = LINK->hap2[system - 1];
  LINK->hap1[system - 1] = j;
  LINK->hap2[system - 1] = i;
  if (system != nlocus)
    hapscr(system + 1, nscramble, LINK);
  else
    setnumbers(LINK);
  LINK->hap1[system - 1] = i;
  LINK->hap2[system - 1] = j;
}  /* hapscr */


static void sethap(system, LINK)
long system;
struct LOC_getlocations *LINK;
{
  long i, j;
  locusvalues *WITH;
  long FORLIM, FORLIM1;

  WITH = thislocus[system - 1];
  if (LINK->thishet[system - 1]) {
    FORLIM = WITH->nallele;
    for (i = 1; i < FORLIM; i++) {
      LINK->hap1[system - 1] = i;
      FORLIM1 = WITH->nallele;
      for (j = i + 1; j <= FORLIM1; j++) {
	LINK->hap2[system - 1] = j;
	if (system != nlocus)
	  sethap(system + 1, LINK);
	else {
	  LINK->rarepresent = checkrare(LINK);
	  if (risk)
	    checkrisk(&LINK->riskhet, &LINK->riskhom, LINK);
	  LINK->there = LINK->start;
	  LINK->thisseg = LINK->here;
	  hapscr(1L, 0L, LINK);
	  LINK->here += LINK->nseg;
	}
      }
    }
    return;
  }
  FORLIM = WITH->nallele;
  for (i = 1; i <= FORLIM; i++) {
    LINK->hap1[system - 1] = i;
    LINK->hap2[system - 1] = i;
    if (system != nlocus)
      sethap(system + 1, LINK);
    else {
      LINK->rarepresent = checkrare(LINK);
      if (risk)
	checkrisk(&LINK->riskhet, &LINK->riskhom, LINK);
      LINK->thisseg = LINK->here;
      LINK->there = LINK->start;
      hapscr(1L, 0L, LINK);
      LINK->here += LINK->nseg;
    }
  }
}  /* sethap */


static void starthap(LINK)
struct LOC_getlocations *LINK;
{
  long i, FORLIM;

  LINK->nseg = 1;
  FORLIM = LINK->nhet;
  for (i = 2; i <= FORLIM; i++)
    LINK->nseg *= 2;
  sethap(1L, LINK);
  LINK->start = LINK->there;
}  /* starthap */


static void gethet1(system, LINK)
long system;
struct LOC_getlocations *LINK;
{
  LINK->thishet[system - 1] = false;
  if (system != nlocus)
    gethet1(system + 1, LINK);
  else
    starthap(LINK);
  LINK->thishet[system - 1] = true;
  LINK->nhet++;
  if (system != nlocus)
    gethet1(system + 1, LINK);
  else
    starthap(LINK);
  LINK->nhet--;
}  /* gethet1 */


void getlocations()
{
  struct LOC_getlocations V;


  V.nhet = 0;
  V.here = 0;
  V.there = 0;
  V.ngene = 0;
  V.start = 0;
  gethet1(1L, &V);
  if (mutsys != 0)
    domutation(&V);
  if (sexlink && risk)
    domalerisk(&V);
}


/* Local variables for inputdata: */
struct LOC_inputdata {
  long nfactor[maxlocus];
} ;

static void inputerror(nerror, par1, par2, LINK)
long nerror, par1, par2;
struct LOC_inputdata *LINK;
{
  printf("Fatal error detected in procedure inputdata\n");
  switch (nerror) {

  case 0:
    printf("Number of loci %2ld exceeds the constant maxlocus\n", par1);
    break;

  case 1:
    printf("Number of loci read %2ld. Less than minimum of 1\n", par1);
    break;

  case 2:
    printf(
      "Error detected reading loci order. Locus number %2ld in position %2ld exceeds number of loci\n",
      par2, par1);
    break;

  case 3:
    printf(
      "Error detected reading loci order. Illegal locus number %2ld in position %2ld\n",
      par2, par1);
    break;

  case 4:
    printf(
      "Error detected reading loci order. Locus number repeated in positions %2ld and %2ld\n",
      par1, par2);
    break;

  case 5:
    printf(
      "Error detected reading locus description. Illegal locus type %2ld for locus %2ld\n",
      par2, par1);
    break;

  case 6:
    printf(
      "Error detected reading locus description for system %2ld. Number of alleles  %2ld exceeds maxall\n",
      par1, par1);
    break;

  case 7:
    printf(
      "Error detected reading locus description for system %2ld. Illegal number of alleles  %2ld\n",
      par1, par2);
    break;

  case 8:
    printf(
      "Error detected reading locus description for system %2ld. Number of factors  %2ld exceeds maxfact\n",
      par1, par2);
    break;

  case 9:
    printf(
      "Error detected reading locus description for system %2ld. Illegal number of factors  %2ld\n",
      par1, par2);
    break;

  case 10:
    printf(
      "Error detected reading locus description for system %2ld. Alleles not codominant\n",
      par1);
    break;

  case 11:
    printf("Error detected reading pedigree record %2ld. Illegal code for sex %2ld\n",
	   par1, par2);
    break;

  case 12:
    printf(
      "Error detected reading pedigree record at pedigree%2ld. Maximum number of pedigree records exceeded\n",
      par1);
    break;

  case 13:
    printf(
      "Error detected reading pedigree record %2ld. Maximum number of individuals exceeded\n",
      par1);
    break;

  case 14:
    printf(
      "Error detected reading pedigree record %2ld. Illegal binary factor code %2ld\n",
      par1, par2);
    break;

  case 15:
    printf(
      "Error detected reading pedigree record %2ld. No allelic pair for genotype\n",
      par1);
    break;

  case 16:
    printf(
      "Error detected reading pedigree record %2ld. Allele number %2ld exceeds maxall\n",
      par1, par2);
    break;

  case 17:
    printf(
      "Error detected reading pedigree record %2ld. Illegal allele number %2ld\n",
      par1, par2);
    break;

  case 18:
    printf("Number of systems after factorization (%3ld) exceeds maxsystem\n",
	   par1);
    break;

  case 19:
    printf("Number of systems after factorization (%3ld) less than minimum of 1\n",
	   par1);
    break;

  case 20:
    printf("Number of recombination types (%3ld) exceeds maxrectype\n", par1);
    break;

  case 21:
    printf("Number of recombination types (%3ld) less than minimum of 1\n",
	   par1);
    break;

  case 22:
    printf(
      "End of file detected in tempdat by procedure readthg before all data found\n");
    break;

  case 23:
    printf(
      "Error detected reading iterated locus in simdata. Value (%3ld) greater than nlocus\n",
      par1);
    break;

  case 24:
    printf("Error detected reading iterated locus in simdata. Illegal value (%3ld)\n",
	   par1);
    break;

  case 25:
    printf("Number of iterated parameters greater then maxn\n");
    break;

  case 26:
    printf(
      "Error detected reading pedigree record %2ld. Liability class (%2ld) exceeds nclass\n",
      par1, par2);
    break;

  case 27:
    printf(
      "Error detected reading pedigree record %2ld. Illegal liability class (%2ld)\n",
      par1, par2);
    break;

  case 28:
    printf(
      "Error detected reading locus description for system%2ld. Liability classes (%3ld) exceed maxliab\n",
      par1, par2);
    break;

  case 29:
    printf(
      "Error detected reading locus description for system%2ld. Illegal number of liability classes (%3ld)\n",
      par1, par2);
    break;

  case 30:
    printf(
      "Error detected reading locus description for system%2ld. Penetrance out of range\n",
      par1);
    break;

  case 31:
    printf(
      "Error detected reading locus description for system%2ld. Number of traits (%3ld) can only be one.\n",
      par1, par2);
    break;

  case 32:
    printf(
      "Error detected reading locus description for system%2ld. Number of traits out of range (%3ld)\n",
      par1, par2);
    break;

  case 33:
    printf(
      "Error detected reading locus description for system%2ld. Variance must be positive\n",
      par1);
    break;

  case 34:
    printf(
      "Error detected reading locus description for system%2ld. Variance multiplier must be positive\n",
      par1);
    break;

  case 35:
    printf(
      "Error detected reading locus description for system%2ld. Risk allele %3ld) exceeds nallele\n",
      par1, par2);
    break;

  case 36:
    printf(
      "Error detected reading locus description for system%2ld. Illegal risk allele (%3ld)\n",
      par1, par2);
    break;

  case 37:
    printf("Error detected reading simdata. Risk locus %3ld) exceeds nlocus\n",
	   par2);
    break;

  case 38:
    printf("Error detected reading simdata. Illegal value for risk locus %3ld)\n",
	   par2);
    break;

  case 39:
    printf("Error detected reading simdata. Mutation locus %3ld) exceeds nlocus\n",
	   par2);
    break;

  case 40:
    printf("Error detected reading simdata. Illegal value for mutation locus %3ld)\n",
	   par2);
    break;
  }
  printf("Press return to halt...\007\n");
  scanf("%*[^\n]");
  getchar();
  exit(0);   /*SUN*/
}  /* inputerror */

/*inputerror*/

static void inputwarning(nwarning, par1, par2, LINK)
long nwarning, par1, par2;
struct LOC_inputdata *LINK;
{
  switch (nwarning) {   /*Version 5.1*/

  case 0:
    printf("Illegal sex difference parameter %2ld Parameter should be 0, 1, or 2\n",
	   par1);
    break;

  case 1:
    printf("Illegal interference parameter %2ld Lack of interference assumed\n",
	   par1);
    break;

  case 2:
    printf(
      "Illegal sex difference parameter %2ld Parameter must be 0 with sex-linked data\n",
      par1);
    break;

  case 3:
    printf("Warning: No risk calculations are permitted during the simulation.\n");
    printf("However the risk locus is %3ld\n", par2);
    break;

  case 4:
    printf(
      "Non-standard affection status%4ld interpreted as normal in pedigree record%5ld\n",
      par2, par1);
    break;
  }
  printf("Press return to continue...\007\n");
  scanf("%*[^\n]");
  getchar();
}  /* inputwarning */

/* Local variables for readped: */
struct LOC_readped {
  struct LOC_inputdata *LINK;
  long sequence;
} ;

/* Local variables for getphenotype: */
struct LOC_getphenotype {
  struct LOC_readped *LINK;
  thisperson **p;
  long system;
} ;

static void readbin(phen, thislocus, LINK)
phenotype **phen;
locusvalues *thislocus;
struct LOC_getphenotype *LINK;
{
  long i, j;
  phenotype *WITH;
  long FORLIM;

  WITH = *phen;
  WITH->which = binary_;
  WITH->phenf = 0;
  FORLIM = LINK->LINK->LINK->nfactor[LINK->system - 1];
  for (i = 1; i <= FORLIM; i++) {
    fscanf(simped, "%ld", &j);
    if (j != 0 && j != 1)
      inputerror(14L, (*LINK->p)->id, j, LINK->LINK->LINK);
    if (j == 1)
      WITH->phenf = ((long)WITH->phenf) | (1L << ((int)i));
  }
}  /* readbin */


static void readnumber(phen, thislocus, LINK, a1, a2)
phenotype **phen;
locusvalues *thislocus;
struct LOC_getphenotype *LINK;
int a1, a2;
{
  long i, j;
  phenotype *WITH;

  WITH = *phen;
  WITH->which = binary_;
  WITH->phenf = 0;
  /*  for (i = 1; i <= 2; i++) {
    fscanf(simped, "%ld", &j);
    if (j > maxall)
      inputerror(16L, (*LINK->p)->id, j, LINK->LINK->LINK);
    if (j < 0)
      inputerror(17L, (*LINK->p)->id, j, LINK->LINK->LINK);
    if (j != 0)
      WITH->phenf = ((long)WITH->phenf) | (1L << ((int)j));
      }*/

  if (a1 > maxall)
    inputerror(16L, (*LINK->p)->id, a1, LINK->LINK->LINK);
  if (a1 < 0)
    inputerror(17L, (*LINK->p)->id, a1, LINK->LINK->LINK);
  if (a2 > maxall)
    inputerror(16L, (*LINK->p)->id, a2, LINK->LINK->LINK);
  if (a2 < 0)
    inputerror(17L, (*LINK->p)->id, a2, LINK->LINK->LINK);
  
  if(a1 != 0)
    WITH->phenf = ((long)WITH->phenf) | (1L << ((int)a1));
  if(a2 != 0)
    WITH->phenf = ((long)WITH->phenf) | (1L << ((int)a2)); 
  
}  /* readnumber */


static void readaff(phen, thislocus, LINK)
phenotype **phen;
locusvalues *thislocus;
struct LOC_getphenotype *LINK;
{
  long thisval;   /*Version 5.1 has this as an integer*/
  phenotype *WITH;

  WITH = *phen;
  WITH->which = affection;
  fscanf(simped, "%ld", &thisval);
  if (thisval == missaff)
    WITH->aff = 0;
  else if (thisval == affval)
    WITH->aff = 2;
  else {
    if (thisval != 1)
      inputwarning(4L, (*LINK->p)->id, thisval, LINK->LINK->LINK);
    WITH->aff = 1;
  }
  if (thislocus->UU.U0.nclass == 1)
    WITH->liability = 1;
  else
    fscanf(simped, "%ld", &WITH->liability);
  if (WITH->liability > thislocus->UU.U0.nclass)
    inputerror(26L, (*LINK->p)->id, WITH->liability, LINK->LINK->LINK);
  if (WITH->liability <= 0)
    inputerror(27L, (*LINK->p)->id, WITH->liability, LINK->LINK->LINK);

  /*Added warning => Version 5.1*/
}  /* readaff */


static void readquan(phen, thislocus, LINK)
phenotype **phen;
locusvalues *thislocus;
struct LOC_getphenotype *LINK;
{
  long i;
  double xval;
  phenotype *WITH;
  long FORLIM;

  WITH = *phen;
  if (!sexlink || !(*LINK->p)->male) {
    WITH->which = quantitative;
    FORLIM = thislocus->UU.U1.ntrait;
    for (i = 0; i < FORLIM; i++)
      fscanf(simped, "%lg", &WITH->x[i]);
    WITH->missing = true;
    FORLIM = thislocus->UU.U1.ntrait;
    for (i = 0; i < FORLIM; i++) {
      if (WITH->x[i] != missval)
	WITH->missing = false;
    }
    return;
  }
  WITH->which = affection;
  fscanf(simped, "%lg", &xval);
  if (xval == missval)
    WITH->aff = missaff;
  else if (xval == affall)
    WITH->aff = affall;
  else
    WITH->aff = unafqu;
  WITH->liability = 1;
  FORLIM = thislocus->UU.U1.ntrait;
  for (i = 2; i <= FORLIM; i++)
    fscanf(simped, "%lg", &xval);
}  /* readquan */

static void getphenotype(p_, LINK, a1, a2)
thisperson **p_;
struct LOC_readped *LINK;
int a1, a2; //allele1 and allele2
{
  struct LOC_getphenotype V;
  long thisread;
  thisperson *WITH;
  long FORLIM;


  V.LINK = LINK;
  V.p = p_;
  WITH = *V.p;
  FORLIM = nlocus;
  for (thisread = 0; thisread < FORLIM; thisread++) {
    V.system = order[thisread];
    WITH->phen[V.system - 1] = NULL;
    if (thislocus[V.system - 1]->which != null_)
      WITH->phen[V.system - 1] = (phenotype *)R_alloc(1,sizeof(phenotype));
    /*         phen[system]:=phenpoint(NewPtr(SizeOf(phenotype)));*/
    switch (thislocus[V.system - 1]->which) {

    case quantitative:
      readquan(&WITH->phen[V.system - 1], thislocus[V.system - 1], &V);
      break;

    case affection:
      readaff(&WITH->phen[V.system - 1], thislocus[V.system - 1], &V);
      break;

    case binary_:
      if (thislocus[V.system - 1]->format == 3)
	readnumber(&WITH->phen[V.system - 1], thislocus[V.system - 1], &V, a1,a2);
      else
	readbin(&WITH->phen[V.system - 1], thislocus[V.system - 1], &V);
      break;
    }
    if (lastpriv == V.system) {
      WITH->privphen = (phenotype *)R_alloc(1,sizeof(phenotype));
      /*          privphen:=phenpoint(NewPtr(SizeOf(phenotype)));*/
      switch (thislocus[V.system - 1]->privlocus->which) {

      case quantitative:
	readquan(&WITH->privphen, thislocus[V.system - 1]->privlocus, &V);
	break;

      case affection:
	readaff(&WITH->privphen, thislocus[V.system - 1]->privlocus, &V);
	break;
      }
    }
  }
}  /* getphenotype */


static void getind(id, LINK)
long *id;
struct LOC_readped *LINK;
{
  long idorig;   /*getind*/

  if (*id == 0) {
    return;
  }  /*Store original ids*/
  idorig = *id;
  *id += LINK->sequence;
  if (*id > maxind)
    inputerror(13L, *id, *id, LINK->LINK);
  if (person[*id] == NULL)
    person[*id] = (thisperson *)R_alloc(1,sizeof(thisperson));
  /*        person[id]:=ind(NewPtr(SizeOf(thisperson))); ;*/
  person[*id]->origid = idorig;
}  /* getind */

/*getind*/

static void multimarriage(p, LINK)
thisperson **p;
struct LOC_readped *LINK;
{
  thisperson *q, *child;   /*multimarriage*/
  thisperson *WITH;

  if ((*p)->foff == NULL) {
    (*p)->multi = false;
    return;
  }
  WITH = *p;
  if (WITH->male)
    q = WITH->foff->ma;
  else
    q = WITH->foff->pa;
  child = WITH->foff;
  (*p)->multi = false;
  do {
    if (WITH->male) {
      WITH->multi = (q == child->ma);
      child = child->nextpa;
    } else {
      WITH->multi = (q == child->pa);
      child = child->nextma;
    }
  } while (!(child == NULL || WITH->multi));
}  /* multimarriage */


static void readped(LINK)
struct LOC_inputdata *LINK;
{
  struct LOC_readped V;
  long i, newid, sex_, profield, newped, nuperson, thisone, thisped;
  long startped[maxped], endped[maxped];
  thisperson *holdloop;
  long FORLIM;
  thisperson *WITH;
  long ii; //individual index for r1.ped

  /*multimarriage*/

  V.LINK = LINK;
  for (i = 0; i <= maxind; i++)
    person[i] = NULL;
  V.sequence = 0;
  nuperson = 0;
  nuped = 1;
  for (i = 0; i < maxloop; i++) {
    looppers[nuped - 1][i][0] = NULL;
    looppers[nuped - 1][i][1] = NULL;
  }
  proband[nuped - 1] = NULL; 
  newped = r1.ped[0][0];
  //  fscanf(simped, "%ld", &newped);
  thisped = newped;
  startped[0] = 1;
  //  while (!P_eof(simped)) { 
  for(ii=0;ii<r1.nind;ii++) {
    nuperson++;
    thisone = r1.ped[ii][1];
    getind(&thisone, &V);
    if (proband[nuped - 1] == NULL)
      proband[nuped - 1] = person[thisone];
    WITH = person[thisone];
    /*               ped := thisped;*/
    WITH->ped = nuped;   /* L. Sandkuijl's suggestion */
    WITH->origped = thisped;   /* L. Sandkuijl's suggestion */
    WITH->id = thisone;
    newid = r1.ped[ii][2];
    getind(&newid,&V);
    WITH->pa = person[newid];
    newid = r1.ped[ii][3];
    getind(&newid,&V);
    WITH->ma = person[newid];
    newid = r1.ped[ii][4];
    getind(&newid,&V);
    WITH->foff = person[newid];
    newid = r1.ped[ii][5];
    getind(&newid,&V);
    WITH->nextpa = person[newid];
    newid = r1.ped[ii][6];
    getind(&newid,&V);
    WITH->nextma = person[newid];
    sex_=r1.ped[ii][7];
    if (sex_ != 1 && sex_ != 2)
      inputerror(11L, WITH->id, sex_, LINK);
    if (sex_ == 1) {
      WITH->male = true;
      men++;
    } else {
      WITH->male = false;
      women++;
    }
    WITH->unknown = false;
    WITH->inloop = 0;
    profield=r1.ped[ii][8];
    person[thisone]->prbnum = profield;
    if (profield == 1)
      proband[nuped - 1] = person[thisone];
    else if (profield > 1 && profield - 1 <= maxloop) {
      if (looppers[nuped - 1][profield - 2][1] == NULL)
	looppers[nuped - 1][profield - 2][1] = person[thisone];
      else
	looppers[nuped - 1][profield - 2][0] = person[thisone];
    }
    getphenotype(&person[thisone], &V, r1.ped[ii][9], r1.ped[ii][10]);
    /* We now require the last column to contain integers: a non-zero
    value in this column indicates that the person is available for
    marker-typing. Learned from Lodewijk Sandkuyl's code.*/
    person[thisone]->availnum=1;
    //    fscanf(simped, "%ld%*[^\n]", &person[thisone]->availnum);
    //    getc(simped);
    WITH = person[thisone];   /*Count the number in each code*/
    if (WITH->availnum != 0 && WITH->availnum != 1 && WITH->availnum != 2 &&
	WITH->availnum != 3) {
      fprintf(simout,
	"ERROR: illegal availability code for person %6ld in pedigree %6ld\n",
	WITH->origid, WITH->origped);
      printf("ERROR: illegal availability code for person %6ld in pedigree %6ld\n",
	     WITH->origid, WITH->origped);
      printf("Press return to halt...\n");
      scanf("%*[^\n]");
      getchar();
      exit(0);   /*SUN*/
    }
    code[WITH->availnum]++;
    if(ii<r1.nind-1)
      newped = r1.ped[ii+1][0];
    if (newped == thisped)
      continue;
    V.sequence += nuperson;
    endped[nuped - 1] = V.sequence;
    nuperson = 0;
    nuped++;
    if (nuped > maxped)
      inputerror(12L, newped, nuped, LINK);
    startped[nuped - 1] = V.sequence + 1;
    for (i = 0; i < maxloop; i++) {
      looppers[nuped - 1][i][0] = NULL;
      looppers[nuped - 1][i][1] = NULL;
    }
    proband[nuped - 1] = NULL;
    thisped = newped;
  }
  totperson = V.sequence + nuperson;
  endped[nuped - 1] = totperson;
  FORLIM = nuped;
  for (newped = 0; newped < FORLIM; newped++) {
    if (looppers[newped][0][1] != NULL && looppers[newped][0][0] == NULL)
      looppers[newped][0][0] = proband[newped];
    for (i = 0; i < maxloop; i++) {
      if (looppers[newped][i][0] == NULL)
	looppers[newped][i][1] = NULL;
      else {
	looppers[newped][i][0]->inloop = i + 1;
	looppers[newped][i][1]->inloop = i + 1;
	if (looppers[newped][i][0]->pa == NULL &&
	    looppers[newped][i][1]->pa != NULL) {
	  holdloop = looppers[newped][i][0];
	  looppers[newped][i][0] = looppers[newped][i][1];
	  looppers[newped][i][1] = holdloop;
	}
      }
    }
  }
  FORLIM = totperson;
  for (thisone = 1; thisone <= FORLIM; thisone++)
    multimarriage(&person[thisone], &V);
}  /* readped */

/* Local variables for readloci: */
struct LOC_readloci {
  struct LOC_inputdata *LINK;
  long i, whichtype, nupriv;
} ;

/* Local variables for getlocus: */
struct LOC_getlocus {
  struct LOC_readloci *LINK;
  long system;
} ;

static void getbin(locus, system, LINK)
locusvalues **locus;
long *system;
struct LOC_getlocus *LINK;
{
  long i, j, k;
  locusvalues *WITH;
  long FORLIM, FORLIM1;

  fscanf(simdata, "%ld%*[^\n]", &LINK->LINK->LINK->nfactor[*system - 1]);
  getc(simdata);
  if (LINK->LINK->LINK->nfactor[*system - 1] > maxfact)
    inputerror(8L, *system, LINK->LINK->LINK->nfactor[*system - 1],
	       LINK->LINK->LINK);
  if (LINK->LINK->LINK->nfactor[*system - 1] <= 0)
    inputerror(9L, *system, LINK->LINK->LINK->nfactor[*system - 1],
	       LINK->LINK->LINK);
  WITH = *locus;
  WITH->UU.U2.numfact = LINK->LINK->LINK->nfactor[*system - 1];
  FORLIM = WITH->nallele;
  for (i = 0; i < FORLIM; i++)
    WITH->UU.U2.allele[i] = 0;
  FORLIM = WITH->nallele;
  for (i = 0; i < FORLIM; i++) {
    FORLIM1 = LINK->LINK->LINK->nfactor[*system - 1];
    for (j = 1; j <= FORLIM1; j++) {
      fscanf(simdata, "%ld", &k);
      if (k == 1)
	WITH->UU.U2.allele[i] = ((long)WITH->UU.U2.allele[i]) | (1L << ((int)j));
    }
  }
  fscanf(simdata, "%*[^\n]");
  getc(simdata);
}  /* getbin */


static void getnumber(locus, system, LINK)
locusvalues **locus;
long *system;
struct LOC_getlocus *LINK;
{
  long i;
  locusvalues *WITH;
  long FORLIM;

  WITH = *locus;
  FORLIM = WITH->nallele;
  for (i = 1; i <= FORLIM; i++) {
    if (i > maxfact) {
      fprintf(simout, " *** ERROR: maxfact not big enough ***\n");
      printf(" *** ERROR: maxfact not big enough ***\n");
      printf(" Press return to halt...\n");
      scanf("%*[^\n]");
      getchar();
      exit(0);   /*SUN*/
    }
    WITH->UU.U2.allele[i - 1] = 1L << ((int)i);
  }
}  /* getnumber */


static void getpen(locus, LINK)
locusvalues **locus;
struct LOC_getlocus *LINK;
{
  long i, j, k, l;
  locusvalues *WITH;
  long FORLIM, FORLIM1, FORLIM2;

  WITH = *locus;
  fscanf(simdata, "%ld%*[^\n]", &WITH->UU.U0.nclass);
  getc(simdata);
  if (WITH->UU.U0.nclass > maxliab)
    inputerror(28L, LINK->system, WITH->UU.U0.nclass, LINK->LINK->LINK);
  if (WITH->UU.U0.nclass <= 0)
    inputerror(29L, LINK->system, WITH->UU.U0.nclass, LINK->LINK->LINK);
  FORLIM = WITH->UU.U0.nclass;
  for (l = 0; l < FORLIM; l++) {
    FORLIM1 = WITH->nallele;
    for (i = 1; i <= FORLIM1; i++) {
      FORLIM2 = WITH->nallele;
      for (j = i - 1; j < FORLIM2; j++) {
	fscanf(simdata, "%lg", &WITH->UU.U0.pen[i][j][2][l]);
	if ((unsigned)WITH->UU.U0.pen[i][j][2][l] > 1.0)
	  inputerror(30L, LINK->system, LINK->system, LINK->LINK->LINK);
	WITH->UU.U0.pen[i][j][1][l] = 1 - WITH->UU.U0.pen[i][j][2][l];
	WITH->UU.U0.pen[i][j][0][l] = 1.0;
	for (k = 0; k <= 2; k++)
	  WITH->UU.U0.pen[j + 1][i - 1][k][l] = WITH->UU.U0.pen[i][j][k][l];
      }
    }
    fscanf(simdata, "%*[^\n]");
    getc(simdata);
    FORLIM1 = WITH->nallele;
    for (i = 0; i < FORLIM1; i++)
      WITH->UU.U0.pen[0][i][0][l] = 1.0;
    if (sexlink) {
      FORLIM1 = WITH->nallele;
      for (i = 0; i < FORLIM1; i++) {   /*Corrected to match Version 5.1*/
	fscanf(simdata, "%lg", &WITH->UU.U0.pen[0][i][2][l]);
	if ((unsigned)WITH->UU.U0.pen[0][i][2][l] > 1.0)
	  inputerror(30L, LINK->system, LINK->system, LINK->LINK->LINK);
      }
      FORLIM1 = WITH->nallele;
      for (i = 0; i < FORLIM1; i++)
	WITH->UU.U0.pen[0][i][1][l] = 1.0 - WITH->UU.U0.pen[0][i][2][l];
      fscanf(simdata, "%*[^\n]");
      getc(simdata);
    }
  }
}  /* getpen */


static void getquan(locus, privelege, LINK)
locusvalues **locus;
boolean privelege;
struct LOC_getlocus *LINK;
{
  /*Get information of a quantitative locus. Privelege says whether it is
  a priveleged locus or not*/
  long i, j, k;
  locusvalues *WITH;
  long FORLIM, FORLIM1, FORLIM2;

  WITH = *locus;
  fscanf(simdata, "%ld%*[^\n]", &WITH->UU.U1.ntrait);
  getc(simdata);
  if (WITH->UU.U1.ntrait > 1)
    inputerror(31L, LINK->system, WITH->UU.U1.ntrait, LINK->LINK->LINK);
  if (WITH->UU.U1.ntrait <= 0)
    inputerror(32L, LINK->system, WITH->UU.U0.nclass, LINK->LINK->LINK);
  FORLIM = WITH->UU.U1.ntrait;
  for (i = 0; i < FORLIM; i++) {
    FORLIM1 = WITH->nallele;
    for (j = 1; j <= FORLIM1; j++) {
      FORLIM2 = WITH->nallele;
      for (k = j; k <= FORLIM2; k++) {
	fscanf(simdata, "%lg", &WITH->UU.U1.pm[j][k - 1][i]);
	WITH->UU.U1.pm[k][j - 1][i] = WITH->UU.U1.pm[j][k - 1][i];
      }
    }
  }
  fscanf(simdata, "%*[^\n]");
  getc(simdata);
  if (privelege && LINK->LINK->nupriv != lastpriv)
    return;
  FORLIM = WITH->UU.U1.ntrait;
  for (i = 0; i < FORLIM; i++) {
    FORLIM1 = WITH->UU.U1.ntrait;
    for (j = i; j < FORLIM1; j++) {
      fscanf(simdata, "%lg", &WITH->UU.U1.vmat[i][j]);
      if (i + 1 == j + 1 && WITH->UU.U1.vmat[i][j] <= 0.0)
	inputerror(33L, LINK->system, LINK->system, LINK->LINK->LINK);
      WITH->UU.U1.vmat[j][i] = WITH->UU.U1.vmat[i][j];
    }
  }
  fscanf(simdata, "%*[^\n]");
  getc(simdata);
  invertmat(WITH->UU.U1.vmat, WITH->UU.U1.ntrait, &WITH->UU.U1.det);
  WITH->UU.U1.det = 1 / sqrt(WITH->UU.U1.det);
  fscanf(simdata, "%lg%*[^\n]", &WITH->UU.U1.conmat);
  getc(simdata);
  if (WITH->UU.U1.conmat <= 0.0)
    inputerror(34L, LINK->system, LINK->system, LINK->LINK->LINK);
  WITH->UU.U1.conmat = 1 / WITH->UU.U1.conmat;
  WITH->UU.U1.contrait = 1.0;
  FORLIM = WITH->UU.U1.ntrait;
  for (i = 1; i <= FORLIM; i++)
    WITH->UU.U1.contrait *= WITH->UU.U1.conmat;
  WITH->UU.U1.contrait = sqrt(WITH->UU.U1.contrait);
}  /* getquan */

static void getlocus(system_, LINK)
long system_;
struct LOC_readloci *LINK;
{
  struct LOC_getlocus V;
  long j;
  locusvalues *WITH, *WITH1;
  long FORLIM;
  int TEMP;


  V.LINK = LINK;
  V.system = system_;
  thislocus[V.system - 1] = (locusvalues *)R_alloc(1,sizeof(locusvalues));
  WITH = thislocus[V.system - 1];
  /*     thislocus[system]:=locuspoint(NewPtr(SizeOf(locusvalues)));*/
  WITH->privlocus = NULL;
  LINK->whichtype = 3;
  WITH->nallele = 2; //SNP marker
  if (LINK->whichtype < 0 && LINK->whichtype > 4)
    inputerror(5L, V.system, LINK->whichtype, LINK->LINK);
  if (WITH->nallele > maxall)
    inputerror(6L, V.system, WITH->nallele, LINK->LINK);
  if (WITH->nallele <= 0)
    inputerror(7L, V.system, WITH->nallele, LINK->LINK);
  switch (LINK->whichtype) {

  case 0:
    WITH->which = quantitative;
    break;

  case 1:
    WITH->which = affection;
    break;

  case 2:
    WITH->which = binary_;
    break;

  case 3:
    WITH->which = binary_;
    break;
  }
  WITH->format = LINK->whichtype;
  if (lastpriv == 0) {
  } else {
    LINK->whichtype = 0;
    if (LINK->whichtype == 0 || LINK->whichtype == 1) {
      LINK->nupriv++;
      WITH->privlocus = (locusvalues *)R_alloc(1,sizeof(locusvalues));
      /*           privlocus:=locuspoint(NewPtr(SizeOf(locusvalues)));*/
      WITH->privlocus->nallele = WITH->nallele;
      WITH1 = WITH->privlocus;
      switch (LINK->whichtype) {

      case 0:
	WITH1->which = quantitative;
	break;

      case 1:
	WITH1->which = affection;
	break;
      }
    }
  }
  if (!disequi) {
    FORLIM = WITH->nallele;
    WITH->freq[0] = r1.af[0];
    WITH->freq[1] = r1.af[1]; //SNP marker
  }
  switch (WITH->which) {

  case binary_:
    if (WITH->format == 3)
      getnumber(&thislocus[V.system - 1], &V.system, &V);
    else
      getbin(&thislocus[V.system - 1], &V.system, &V);
    break;

  case affection:
    getpen(&thislocus[V.system - 1], &V);
    break;

  case quantitative:
    getquan(&thislocus[V.system - 1], false, &V);
    break;
  }
  if (WITH->privlocus != NULL) {
    switch (WITH->privlocus->which) {

    case affection:
      getpen(&WITH->privlocus, &V);
      break;

    case quantitative:
      getquan(&WITH->privlocus, true, &V);
      break;
    }
  }
  if (LINK->nupriv == lastpriv && lastpriv != 0)
    lastpriv = V.system;
  if (!risk)
    return;
  /*  if (LINK->i == risksys) {
    fscanf(simdata, "%d%*[^\n]", &TEMP);
    getc(simdata);
    riskall = TEMP;
  }
  if (riskall > thislocus[LINK->i - 1]->nallele)
    inputerror(35L, LINK->i, (long)riskall, LINK->LINK);
  if (riskall < 0)
  inputerror(36L, LINK->i, (long)riskall, LINK->LINK); */
}  /* getlocus */


static void gettheta(sex_, LINK)
thetavalues **sex_;
struct LOC_readloci *LINK;
{
  thetarray oldtheta;
  long i;
  thetavalues *WITH;
  long FORLIM;

  *sex_ = (thetavalues *)R_alloc(1,sizeof(thetavalues));
  WITH = *sex_;
  /*     sex:=thetapoint(NewPtr(SizeOf(thetavalues)));*/
  if (*sex_ == maletheta || readfemale) {
    FORLIM = nlocus - 2;
    /*    for (i = 0; i <= FORLIM; i++)
      fscanf(simdata, "%lg", &WITH->theta[i]);
    if (interfer && !mapping)
      fscanf(simdata, "%lg", &WITH->theta[nlocus - 1]);
    fscanf(simdata, "%*[^\n]");
    getc(simdata);*/
  } else {
    /*    fscanf(simdata, "%lg%*[^\n]", &distratio);
	  getc(simdata);*/
  }
  /*         FOR j:=1 TO maxneed DO segprob[j]:=0.0;*/
  if (!interfer || mapping) {
    return;
  }  /*=ln(1/0.0001-1.0)*/
  memcpy(oldtheta, WITH->theta, sizeof(thetarray));
  WITH->theta[0] = (oldtheta[0] + oldtheta[nlocus - 1] - oldtheta[nlocus - 2]) / 2.0;
  WITH->theta[nlocus - 2] =
    (oldtheta[nlocus - 2] + oldtheta[nlocus - 1] - oldtheta[0]) / 2.0;
  WITH->theta[nlocus - 1] =
    (oldtheta[0] + oldtheta[nlocus - 2] - oldtheta[nlocus - 1]) / 2.0;
  FORLIM = nlocus;
  for (i = 0; i < FORLIM; i++) {
    if (WITH->theta[i] > 0.0) {
      if (WITH->theta[i] < 0.999)
	WITH->theta[i] = log(1.0 / WITH->theta[i] - 1.0);
      else
	WITH->theta[i] = -6.9;   /*=ln(1/0.999-1.0)*/
    } else
      WITH->theta[i] = 9.21;
  }
}  /* gettheta */

/*readped*/

static void readloci(LINK)
struct LOC_inputdata *LINK;
{
  struct LOC_readloci V;
  long j, coupling, autosomal, independent, difference, FORLIM, FORLIM1;
  locusvalues *WITH;


  V.LINK = LINK;
  lastpriv = 0;
  nlocus = r1.nlocus;
  risksys = 0;
  autosomal = r1.autosomal;
  if (nlocus > maxlocus)
    inputerror(0L, nlocus, nlocus, LINK);
  if (nlocus <= 0)
    inputerror(1L, nlocus, nlocus, LINK);
  if (risksys > maxlocus)
    inputerror(37L, risksys, risksys, LINK);
  if (risksys < 0)
    inputerror(38L, risksys, risksys, LINK);
  if (risksys != 0)
    inputwarning(3L, risksys, risksys, LINK);
  risk = false;   /*risksys<>0*/
  sexlink = (autosomal == 1);
#if 0
  printf("YOU ARE USING LINKAGE/SLINK (V%s) WITH%3ld-POINT", version, nlocus);
  if (sexlink)
    printf(" SEXLINKED DATA\n");
  else
    printf(" AUTOSOMAL DATA\n");
#endif
  mutsys=mutmale=mutfemale=coupling=0;
  if (mutsys > maxlocus)
    inputerror(39L, mutsys, mutsys, LINK);
  if (mutsys < 0)
    inputerror(40L, mutsys, mutsys, LINK);
  if (coupling == 1)
    disequi = true;
  else
    disequi = false;
  if (disequi)
    hapfreq = (thisarray *)R_alloc(1,sizeof(thisarray));
  FORLIM = nlocus;
  /*     hapfreq:=genpoint(NewPtr(SizeOf(thisarray)));*/
  /*for (V.i = 1; V.i <= FORLIM; V.i++) {
    fscanf(simdata, "%ld", &j);
    if (j > nlocus)
      inputerror(2L, V.i, j, LINK);
    if (j <= 0)
      inputerror(3L, V.i, j, LINK);
    order[j - 1] = V.i;
    } */
  order[0]=1; //single locus
  FORLIM = nlocus;
  for (V.i = 1; V.i <= FORLIM; V.i++) {
    FORLIM1 = V.i;
    for (j = 1; j < FORLIM1; j++) {
      if (order[V.i - 1] == order[j - 1])
	inputerror(4L, V.i, j, LINK);
    }
  }
  if (mutsys != 0)
    mutsys = order[mutsys - 1];
  V.nupriv = 0;
  FORLIM = nlocus;
  for (V.i = 1; V.i <= FORLIM; V.i++)
    getlocus(order[V.i - 1], &V);
  if (risksys != 0)
    risksys = order[risksys - 1];
  increment[nlocus - 1] = 1;
  for (V.i = nlocus - 1; V.i >= 1; V.i--)
    increment[V.i - 1] = increment[V.i] * thislocus[V.i]->nallele;
  fgeno = 1;
  FORLIM = nlocus;
  for (j = 0; j < FORLIM; j++)
    fgeno *= thislocus[j]->nallele;
  mgeno = fgeno;
  nuhap = fgeno;
  FORLIM = nlocus;
  for (V.i = 1; V.i <= FORLIM; V.i++)
    nohom[V.i - 1] = false;
  /*  if (disequi) {
    FORLIM = mgeno;
    for (V.i = 1; V.i <= FORLIM; V.i++)
      fscanf(simdata, "%lg", &hapfreq->genarray[V.i - 1]);
    fscanf(simdata, "%*[^\n]");
    getc(simdata);
    } else { */
    FORLIM = nlocus;
    for (V.i = 1; V.i <= FORLIM; V.i++) {
      WITH = thislocus[V.i - 1];
      if (WITH->which == affection || WITH->which == quantitative) {
	if (WITH->freq[affall - 1] < minfreq)
	  nohom[V.i - 1] = true;
      }
    }
    /* } */
  fgeno = fgeno * (fgeno + 1) / 2;
  if (!sexlink)
    mgeno = fgeno;
  difference = 0;
  if ((unsigned long)difference > 2) {
    inputwarning(0L, difference, difference, LINK);
    difference = 0;
  }
  sexdif = (difference != 0);
  readfemale = (difference == 2);
  independent = 0;
  if ((unsigned long)independent > 2) {
    inputwarning(1L, independent, independent, LINK);
    independent = 0;
  }
  interfer = (independent != 0);
  mapping = (independent == 2);
  gettheta(&maletheta, &V);
  if (sexdif)
    gettheta(&femaletheta, &V);
  else
    femaletheta = maletheta;
  if (sexlink && sexdif) {
    inputwarning(2L, difference, difference, LINK);
    sexdif = false;
    readfemale = false;
  }
  //parameters not in use
  which=0;
  inc=0.1;
  finish=0.2;
  finish += 0.0001;
  if (!sexlink) {
    if (mutsys == 0)
      thispath = auto_;
    else
      thispath = mauto;
  } else if (mutsys == 0)
    thispath = sex;
  else
    thispath = msex;
  segscale = scale;
  FORLIM = nlocus;
}  /* readloci */


void inputdata()
{
  struct LOC_inputdata V;

  readloci(&V);
  readped(&V);
}  /* inputdata */



