/* Output from p2c, the Pascal-to-C translator */
/* From input file "unknown.p" */

/*Changes at Columbia U. indicated by "change"*/
/*10 July 1993*/


/*
   Modified extensively by Dylan Cooper and Alejandro Schaffer
   to produce a file whose name
   is in the macro LOOPFILE_NAME) that is used to restrict the
   genotypes of unknowns in the presence of loops.  Modified in late
   1994 and summer of 1995.
 
   Please read README.loopfile for a description of the file created
   and the concept of loopbreaker vectors.
 
   The basic idea is (for each locus of each pedigree):
       -- find all the permutations (vectors) of single locus genotypes
          that the loop breakers can have
       -- traverse the pedigree from the proband
          for each loopbreaker vector and record the single locus
          genotypes each person may have if the loopbreakers have the
          single locus genotypes specified by the vector
       -- if there are no loops do a faster version of the old alg

  Modified by Alejandro Schaffer in late 1995 to renumber alleles when
  ALLELE_SPEED is set to 1. If 2 or more alleles are unused in a pedigree
  at a locus, all the unused alleles can be combined into 1.

  Allele renumbering affects the output of speedfile.dat and loopfile.dat,
  but not pedfile.dat, datafile.dat, or ipedfile.dat. The reasons for this
  choice are somewhat arbitrary. 


  Further changes were made in April,1996 by Tony Schurtz, in order to
  allow for backward compatibility with the old "speedfile.dat" format.
  Now the application always writes out speedfile.dat in the original
  file format (as though ALLELE_SPEED were zero).  If ALLELE_SPEED is set
  to one, a second speedfile is written, in the new format, to a file called
  "newspeedfile.dat"

  The changes to the source are quite systematic.  Each function which
  depended upon the value of ALLELE_SPEED was simply "cloned", one version
  retaining the ALLELE_SPEED = 0 code, and the other the ALELE_SPEED = 1
  code.  These were named x_old, and x_new, respectively, x being the old
  function name..  main() was cloned in the same way in order to ensure
  that the entire state of the application is reinitialized before
  newspeedfile.dat is generated.
  Some of these routines were merged by A. A. Schaffer in February 1997.
*/

#include "unknown.h"

static subhap seghap[maxgeno];
static thisperson *person[maxind + 1];
static thisperson *proband, *loop1, *loop2;
static long risksys, mutsys;
static long fgeno, mgeno;
static long nsequence, newped, whichsys, totperson;
static boolean sexlink, risk, disequi;
static boolean disfreqs; /*using conditional allele frequencies, Morgan*/
static genotype gene;
static double one;   /*changed*/
static boolean makehomozygous;   /* Change - Added 7/8/93 */
static int numind;  /*number of individuals in pedigree*/
static int depth;  /*depth of recursion*/

/*The following three variables were added by A. A. Schaffer to
 help pinpoint Mendelian incompatibilities*/
static boolean incompat_traversal; /*found incompatibility on this traversal*/
static boolean first_incompat; /*Is this the first incomptibility */
static boolean detected_incompat[maxind+1]; /* array to record whether
                                               an incompatibility with this
                                               person has already been
                                               reported*/
static boolean errorOccurred; /*is there a basic error in the input files*/
static boolean gapsinIds; /*is there a gap in the individual ids*/
static int totalNumPed; /*total number of pedigrees*/

static long genenumber[maxgeno][maxgeno];

/* table holding possible genotypes for unknowns.  Indexed by:
   - person[i]->nseq
   - locus
   - loopbreaker vector #
   - single locus genotype #
*/
static geno_for_unknown *unknown_poss = NULL;

/* array of valid loopbreakers */
static loop_vector_array loop_vectors = NULL;

/* holds number of loopbreaker vectors for each locus */
static int num_loop_vectors[maxlocus];

char status[maxind+1];                    /* used to check pedfile.dat
                                            for consistency*/

/* ----------------------- Procedures ------------------------------- */



/* Local variables for readloci: */
struct LOC_readloci {
  long whichtype;
} ;

/* Local variables for getlocus: */
struct LOC_getlocus {
  struct LOC_readloci *LINK;
  long system;
} ;

void getnumber(locus, LINK)
locusvalues **locus;
struct LOC_getlocus *LINK;
{
  long i;
  locusvalues *WITH;
  long FORLIM;

  WITH = *locus;
  FORLIM = WITH->nallele;
  for (i = 1; i <= FORLIM; i++)
    WITH->UU.U2.allele[i - 1] = 1;
}  /*getnumber*/


void getlocus(system_, LINK)
long system_;
struct LOC_readloci *LINK;
{
  struct LOC_getlocus V;
  locusvalues *WITH;

  V.LINK = LINK;
  V.system = system_;
  thislocus[V.system - 1] = (locusvalues *) R_alloc(1,sizeof(locusvalues));
  WITH = thislocus[V.system - 1];
  //fscanf(datafile, "%ld%ld", &LINK->whichtype, &WITH->nallele);
  LINK->whichtype = 3;
  WITH->nallele = 2; //SNP marker
  WITH->maxallele = WITH->nallele;
#if LOOPSPEED
  WITH->fgeno = (WITH->nallele * (WITH->nallele + 1) / 2);
  if (sexlink)
    WITH->mgeno = WITH->nallele;
  else
    WITH->mgeno = WITH->fgeno;
#endif
  switch (LINK->whichtype) {

  case 0:
    WITH->which = quantitative;
    break;

  case 1:
    WITH->which = affection;
    break;

  case 2:
    WITH->which = binary_;
    WITH->UU.U2.format = binformat;
    break;

  case 3:
    WITH->which = binary_;
    WITH->UU.U2.format = allformat;
    break;
  }
  if (!disequi) {
  //  fscanf(datafile, "%*[^\n]");    
  //  fscanf(datafile, "%lf", &thislocus[V.system - 1]->onefreq);
    thislocus[V.system - 1]->onefreq = r1.af[0];
  }
  //else {
  //fscanf(datafile, "%*[^\n]");
  //getc(datafile);
  // }
  //if (!disequi) {
  //  fscanf(datafile, "%*[^\n]");
  //  getc(datafile);
  //} 
  //if (disfreqs && (binary_==WITH->which)) {
  //  fscanf(datafile, "%*[^\n]");
  //  getc(datafile);
  //}

  switch (WITH->which) {

  case quantitative:
    //getquan(&thislocus[V.system - 1], &V);
    break;

  case affection:
    //getpen(&thislocus[V.system - 1], &V);
    break;

  case binary_:
    getnumber(&thislocus[V.system - 1], &V);
    break;
  }
}  /*getlocus*/



static void readloci()
{
  struct LOC_readloci V;
  long coupling, autosomal;
  double mu;
  int i;

  //fscanf(datafile, "%ld%ld%ld%*[^\n]", &nsystem, &risksys, &autosomal);
  nsystem = r1.nlocus;
  risksys = 0;
  autosomal = r1.autosomal;
  //getc(datafile);
  //if (nsystem > maxlocus)
  //  inputerror(0L, nsystem, nsystem);
  //if (nsystem <= 0)
  //  inputerror(1L, nsystem, nsystem);
  risk = (risksys != 0);
  sexlink = (autosomal == 1);
  //fscanf(datafile, "%ld%lf%lf%ld%*[^\n]", &mutsys, &mu, &mu, &coupling);
  mutsys = mu = coupling = 0;
  //getc(datafile);
  disequi = (coupling == 1);

  //fscanf(datafile, "%*[^\n]");
  //getc(datafile);
  for (i = 1; i <= nsystem; i++) 
    getlocus((long) i, &V);
  
  if(disequi && disfreqs) {

    fprintf(stderr,"\nYou cannot use the disequilibrium model with conditional allele frequencies.");
    exit(EXIT_FAILURE);
  }
}  /*readloci*/

void initial_breakers(int seed)
{
  boolean *islooped; /*which pedigrees are looped*/
  int i; /*index over pedigrees*/

  readloci();
  gapsinIds = FALSE;
  totalNumPed = r1.nped;   
  islooped = (boolean *) R_alloc(totalNumPed, sizeof(boolean));
  for(i = 0; i < totalNumPed; i++)
    islooped[i] = FALSE;
  detectLoopedPedigrees(islooped);
  loopbreakers2(islooped,seed);
}


extern void loopbreakerc(int *nind_R, int *nped_R, int *nlocus_R, int *ped_R, 
		   int *autosomal_R, double *af_R, int *seed_R, 
		   int *trait_R, int *hetvalue_R,
		   int *pedout_R) 
{ 
  int i,j;
  
  //r1: input information from R
  //r2: output pedigrees

  r1.nind = *nind_R;
  r1.nlocus = *nlocus_R;
  
  r1.ped = (int **) R_alloc(r1.nind,sizeof(int *));
  for(i=0;i<r1.nind;i++) {
    r1.ped[i] = (int *) R_alloc(r1.nlocus*2+10,sizeof(int));
    for (j=0;j<r1.nlocus*2+10;j++) {
      r1.ped[i][j] = ped_R[i+j*r1.nind];
    }
  } //change *ped_R into a matrix 
    //the format of matrix in R changes row index first
  
  r1.autosomal = *autosomal_R;
  r1.af = af_R;
  r1.nped = *nped_R;
  r1.seed = *seed_R;
  r1.trait = *trait_R;
  r1.hetvalue = *hetvalue_R;

  r2.nind = r1.nind*2; //upper limit of nind for the output
  r2.nlocus = r1.nlocus;

  r2.ped = (int **) R_alloc(r2.nind,sizeof(int *));
  for (i=0;i<r2.nind;i++) 
    r2.ped[i] = (int *) R_alloc(r2.nlocus*2+10,sizeof(int));

  initial_breakers(r1.seed);

  //transform matrix mc1.ped into a vector in R
  for (i=0;i<r2.nind;i++) 
    for (j=0;j<r2.nlocus*2+10;j++) 
      pedout_R[i+j*r1.nind*2] = r2.ped[i][j];
  
}
